#! /usr/bin/env python
# ==========================================================================
# Compute a visibility cube
#
# Copyright (C) 2016-2019 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import math
import gammalib
import ctools


# =============== #
# csviscube class #
# =============== #
class csviscube(ctools.cscript):
    """
    Compute a visibility cube
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor

        Parameters
        ----------
        argv : list of str
            List of IRAF command line parameter strings of the form
            ``parameter=3``.

        Raises
        ------
        TypeError
            An invalid number of command line arguments was provided.
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Initialise members
        self._cube    = gammalib.GSkyMap()
        self._results = []

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get all parameters
        """
        # Query parameters
        self['mjdref'].real()
        self['tmin'].time()
        self['tmax'].time()
        self['geolon'].real()
        self['geolat'].real()
        self['outfile'].filename()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return

    def _zenith_angle_map(self, nx, ny, dx, dy):
        """
        Compute zenith angle map

        The zenith angle of a position (ra,dec) depends on the declination and
        the hour angle h and is given by

        zenith(h,dec) = arccos( sin(lat) * sin(dec) + cos(lat) * cos(dec) * cos(h) )

        The hour angle h (or local hour angle, LHA) is defined as the difference
        between local siderial time (LST) and the Right Ascension

        h = LST - ra

        The map is computed for h=-ra which is equivalent to GST=lon (or LST=0).
        In other words, the map corresponds to the time when ra=0 passes through
        the local meridian.

        Parameters
        ----------
        nx : int
            Number of Right Ascension pixels.
        ny : int
            Number of Declination pixels.
        dx : float
            Right Ascension pixel size in degrees.
        dy : float
            Declination pixel size in degrees.

        Returns
        -------
        zmap : `~gammalib.GSkyMap`
            Allsky map comprising the zenith angle for an hour angle of 0.
        """
        # Initialise zenith angle map
        zmap = gammalib.GSkyMap('CAR','CEL',0.0,0.0,-dx,dy,nx,ny)

        # Set hour angle and declination vectors
        hours = [(float(i)+0.5)*dx      for i in range(nx)]
        decs  = [(float(i)+0.5)*dy-90.0 for i in range(ny)]

        # Get array geographic latitude
        geolat = self['geolat'].real()

        # Precompute latitude terms
        cos_lat = math.cos(geolat*gammalib.deg2rad)
        sin_lat = math.sin(geolat*gammalib.deg2rad)

        # Loop over all declinations and hour angles and compute the zenith
        # angle. Store the zenith angle in the map
        index = 0
        for dec in decs:
            cos_dec = math.cos(dec*gammalib.deg2rad)
            sin_dec = math.sin(dec*gammalib.deg2rad)
            for h in hours:
                cos_h  = math.cos(h*gammalib.deg2rad)
                zenith = math.acos(sin_lat*sin_dec +
                                   cos_lat*cos_dec*cos_h)*gammalib.rad2deg
                zmap[index] = zenith
                index      += 1

        # Log zenith angle map
        self._log_header2(gammalib.EXPLICIT, 'Zenith angle map')
        self._log_string(gammalib.EXPLICIT, str(zmap))

        # Return zenith angle map
        return zmap

    def _sun_ra_exclusion(self, time):
        """
        Compute the half length of the Right Ascension interval to exclude due
        to the Sun constraint

        The Sun zenith angle constraint implies an interval in Right Ascension
        that is to be excluded (that's the interval during which it is day).
        This method computes half the length of this interval in degrees.

        Parameters
        ----------
        time : `~gammalib.GTime()`
            Time for which the Sun exclusion is to be computed

        Returns
        -------
        dra : float
            Right Ascension interval half length (degrees)
        """
        # Get array geographic latitude and minimum Sun zenith angle in radians
        geolat    = self['geolat'].real()    * gammalib.deg2rad
        sunzenith = self['sunzenith'].real() * gammalib.deg2rad

        # Compute Right Ascension and Declination of Sun in degrees and
        # convert the Declination into radians
        sun = gammalib.GSkyDir()
        sun.sun(time)
        sundec = sun.dec()

        # Compute some sines and cosines
        cos_sunzenith = math.cos(sunzenith)
        sin_geolat    = math.sin(geolat)
        cos_geolat    = math.cos(geolat)
        sin_sundec    = math.sin(sundec)
        cos_sundec    = math.cos(sundec)

        # Compute Right Ascension difference when Sun is below the mimimum
        # zenith angle in degrees
        dra   = 0.0
        nom   = cos_sunzenith - sin_geolat * sin_sundec
        denom = cos_geolat * cos_sundec
        if denom != 0.0:
            arg = nom/denom
            if arg >= -1.0 and arg <= 1.0:
                dra = math.acos(arg) * gammalib.rad2deg

        # Return
        return dra

    def _moon_ra_exclusion(self, time):
        """
        Compute the half length of the Right Ascension interval to exclude due
        to the Moon constraint
        
        The Moon zenith angle constraint implies an interval in Right Ascension
        that is to be excluded. This method computes half the length of this
        interval.

        Parameters
        ----------
        time : `~gammalib.GTime()`
            Time for which the Moon exclusion is to be computed

        Returns
        -------
        dra : float
            Right Ascension interval half length (degrees)
        """
        # Get array geographic latitude and minimum Moon zenith angle in radians
        geolat     = self['geolat'].real()     * gammalib.deg2rad
        moonzenith = self['moonzenith'].real() * gammalib.deg2rad

        # Compute Right Ascension and Declination of Moon in degrees and
        # convert the Declination into radians
        moon = gammalib.GSkyDir()
        moon.moon(time)
        moondec = moon.dec()

        # Compute some sines and cosines
        cos_moonzenith = math.cos(moonzenith)
        sin_geolat     = math.sin(geolat)
        cos_geolat     = math.cos(geolat)
        sin_moondec    = math.sin(moondec)
        cos_moondec    = math.cos(moondec)

        # Compute Right Ascension difference when Moon is below the mimimum
        # zenith angle in degrees
        dra   = 0.0
        nom   = cos_moonzenith - sin_geolat * sin_moondec
        denom = cos_geolat * cos_moondec
        if denom != 0.0:
            arg = nom/denom
            if arg >= -1.0 and arg <= 1.0:
                dra = math.acos(arg) * gammalib.rad2deg

        # Return
        return dra

    def _adjust_ra_interval(self, ra_start, ra_stop):
        """
        Adjust Right Ascension interval so that it overlaps with the [0,360[
        interval

        Parameters
        ----------
        ra_start : float
            Start of interval to exclude (degrees)
        ra_stop : float
            Stop of interval to exclude (degrees)

        Returns
        -------
        ra_start : float
            Adjusted start of interval to exclude (degrees)
        ra_stop : float
            Adjusted stop of interval to exclude (degrees)
        """
        # Adjust interval if both boundaries are negative
        while ra_start < 0.0 and ra_stop < 0.0:
            ra_start += 360.0
            ra_stop  += 360.0

        # Adjust interval if both boundaries are equal or larger than 360
        # degrees
        while ra_start >= 360.0 and ra_stop >= 360.0:
            ra_start -= 360.0
            ra_stop  -= 360.0

        # Return interval
        return ra_start, ra_stop

    def _exclude_ra_interval(self, hours, ra_start, ra_stop):
        """
        Exclude Right Ascension interval from array of hour angles

        Parameters
        ----------
        hours : list of floats
            Array of hours
        ra_start : float
            Start of interval to exclude (degrees)
        ra_stop : float
            Stop of interval to exclude (degrees)

        Returns
        -------
        hours : list of floats
            Array of hours with excluded Right Ascension interval
        """
        # Conversion from RA (degrees) to index
        nhours = len(hours)
        ra2inx = float(nhours)/360.0

        # Adjust interval
        ra_start, ra_stop = self._adjust_ra_interval(ra_start, ra_stop)

        # Case 1: The RA interval is fully comprised within the [0,360]
        #         interval, hence we simply exclude this interval.
        if ra_start >= 0.0 and ra_stop <= 360.0:
            inx_start = int(ra_start * ra2inx + 0.5)
            inx_stop  = int(ra_stop  * ra2inx + 0.5)
            for i in range(inx_start,inx_stop):
                hours[i] = 0.0

        # Case 2: The RA interval is starting at negative RA. In that case
        #         there are two intervals to exclude:
        #         [ra_start+360,360] and [0,ra_stop]
        elif ra_start < 0.0:
            inx_start = int((ra_start+360.0) * ra2inx + 0.5)
            inx_stop  = int(ra_stop * ra2inx + 0.5)
            for i in range(inx_start,nhours):
                hours[i] = 0.0
            for i in range(0,inx_stop):
                hours[i] = 0.0

        # Case 3: The RA interval is stopping at RA>360. In that case there
        #         are two intervals to exclude:
        #         [0,ra_stop-360] and [0,ra_start]
        elif ra_stop > 360.0:
            inx_start = int(ra_start * ra2inx + 0.5)
            inx_stop  = int((ra_stop-360.0) * ra2inx + 0.5)
            for i in range(0,inx_stop):
                hours[i] = 0.0
            for i in range(inx_start,nhours):
                hours[i] = 0.0

        # Return hours
        return hours

    def _hour_angle_weight(self):
        """
        Compute hour angle weights

        Computes an array that specifies the time during which a given hour
        angle was observed during the observing time interval [tmin,tmax].
        The hour angle runs from 0 to 360 degrees, array values are in units
        of hours.

        The method loops over the days in the time interval [tmin,tmax], with
        time = tmin + i*86400 seconds. At each time step the position in Right
        Ascension and Declination of the Sun is computed (the Sun's
        Declination is in fact not used here).

        An interval of [ra_sun-dra_sun, ra_sun+dra_sun] is assumed to be the
        day. The method _sun_ra_exclusion() is used to compute dra_sun, which
        is half of the length of the day in degrees (recall that 15 degrees
        is one hour). The length of the day is computed using the sunzenith
        constraint.
        """
        # Write header
        self._log_header2(gammalib.NORMAL, 'Hour angle weights')

        # Get MET time reference
        tref = gammalib.GTimeReference(self['mjdref'].real(),'s','TT','LOCAL')

        # Get time interval
        tmin = self['tmin'].time(tref)
        tmax = self['tmax'].time(tref)

        # Initialise hour angle list and results
        hour_angles   = []
        self._results = []

        # Set number of hour angle bins and compute conversion factor and weight
        nsteps = 1000
        ra2inx = float(nsteps)/360.0 # Conversion from RA (deg) to index
        weight = 24.0/float(nsteps)  # Weight per hour angle step

        # Initialise hour angle array and weight
        hours  = [0.0 for i in range(nsteps)]

        # Initialise time and loop until the end time is reached
        time = tmin
        while time <= tmax:

            # Initialise hours array for this time step
            hours_time = [weight for i in range(nsteps)]

            # Compute half the length of the exclusion intervals in degrees
            sun_dra  = self._sun_ra_exclusion(time)
            moon_dra = self._moon_ra_exclusion(time)

            # Compute Right Ascension and Declination of Sun in degrees and
            # derive Sun exclusion interval
            sun = gammalib.GSkyDir()
            sun.sun(time)
            sun_ra       = sun.ra_deg()
            sun_dec      = sun.dec_deg()
            sun_ra_start = sun_ra - sun_dra
            sun_ra_stop  = sun_ra + sun_dra

            # Adjust interval
            sun_ra_start, sun_ra_stop = self._adjust_ra_interval(sun_ra_start,
                                                                 sun_ra_stop)

            # Compute Right Ascension and Declination of Moon in degrees and
            # derive Moon exclusion interval
            moon = gammalib.GSkyDir()
            moon.moon(time)
            moon_ra       = moon.ra_deg()
            moon_dec      = moon.dec_deg()
            moon_ra_start = moon_ra - moon_dra
            moon_ra_stop  = moon_ra + moon_dra

            # Adjust interval
            moon_ra_start, moon_ra_stop = self._adjust_ra_interval(moon_ra_start,
                                                                   moon_ra_stop)

            # Compute Moon elongation and illumation fraction (Moon phase)
            elongation = sun.dist_deg(moon)
            fli        = (1.0 - math.cos(elongation * gammalib.deg2rad))/2.0

            # Exclude hours due to Sun constraint (this defines the night)
            hours_time = self._exclude_ra_interval(hours_time, sun_ra_start,
                                                               sun_ra_stop)

            # Exclude hours due to Moon constraint if the illumination fraction
            # is equal to or above the maximim fraction of illumination
            if fli >= self['maxfli'].real():
                hours_time = self._exclude_ra_interval(hours_time, moon_ra_start,
                                                                   moon_ra_stop)

            # Add hours
            dark_time = 0.0
            for i in range(nsteps):
                hours[i]  += hours_time[i]
                dark_time += hours_time[i]

            # Set result record
            result = {'time': time.copy(),
                      'sun_ra': sun_ra, 'sun_dec': sun_dec,
                      'moon_ra': moon_ra, 'moon_dec': moon_dec,
                      'elongation': elongation, 'fli': fli,
                      'sun_ra_start': sun_ra_start, 'sun_ra_stop': sun_ra_stop,
                      'moon_ra_start': moon_ra_start, 'moon_ra_stop': moon_ra_stop,
                      'dark_time' : dark_time}

            # Append result record to results
            self._results.append(result)

            # Log results
            logs = 'Sun=(%8.3f,%7.3f) Moon=(%8.3f,%7.3f) '\
                   'Sun_RA_excl=[%8.3f,%8.3f] Moon_RA_excl=[%8.3f,%8.3f] '\
                   'FLI=%4.2f Dark time=%5.2f h' % \
                   (sun_ra, sun_dec, moon_ra, moon_dec,
                    sun_ra_start, sun_ra_stop, moon_ra_start, moon_ra_stop,
                    fli, dark_time)
            self._log_value(gammalib.VERBOSE, time.utc(), logs)

            # Add seconds of one day and start with next day
            time += 86400.0

        # Build hour angle list
        dh = 360.0/float(nsteps)
        for i in range(nsteps):
            hour_angle = {'angle': dh*float(i), 'hours': hours[i]}
            hour_angles.append(hour_angle)

        # Log hour angle weights
        total = 0.0
        for hour_angle in hour_angles:
            total += hour_angle['hours']
        self._log_value(gammalib.EXPLICIT, 'Observing time', str(total)+' h')
        self._log_value(gammalib.EXPLICIT, 'Number of hour angle bins', len(hour_angles))

        # Return hour angle weights
        return hour_angles

    def _visibility_cube(self):
        """
        Compute visibility cube

        Compute visibility cube by displacing the zenith angle map for all
        hour angles. The visibility cube contains the number of hours a given
        celestial position is visible under a given zenith angle interval.
        Summing over all zenith angle intervals specifies for how long a given
        celestial position will be visible.
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Compute visibility cube')

        # Set visibility cube and zenith angle map dimensions and bin size
        binsz = self['binsz'].real()
        nx    = int(360.0/binsz+0.5)
        ny    = int(180.0/binsz+0.5)
        dx    = 360.0/float(nx)
        dy    = 180.0/float(ny)

        # Set zenith angle dimension and bin size
        dz   = self['dz'].real()
        zmax = self['zmax'].real()
        nz   = int(zmax/dz+0.5)
        dz   = zmax/float(nz)

        # Log dimensions
        self._log_header2(gammalib.NORMAL, 'Visibility cube dimensions')
        self._log_value(gammalib.NORMAL, 'Number of RA bins', nx)
        self._log_value(gammalib.NORMAL, 'Number of Dec bins', ny)
        self._log_value(gammalib.NORMAL, 'Number of zenith angles', nz)
        self._log_value(gammalib.NORMAL, 'RA pixel size', str(dx)+' deg')
        self._log_value(gammalib.NORMAL, 'Dec pixel size', str(dy)+' deg')
        self._log_value(gammalib.NORMAL, 'Zenith angle bin size', str(dz)+' deg')
        self._log_value(gammalib.NORMAL, 'Maximum zenith angle', str(zmax)+' deg')

        # Compute zenith angle map
        zmap = self._zenith_angle_map(nx,ny,dx,dy)

        # Setup hour angle weights. They specify for how many hours a given
        # hour angle is observed during the covered time period.
        hour_angles = self._hour_angle_weight()

        # Initialise visibility cube
        self._cube = gammalib.GSkyMap('CAR','CEL',0.0,0.0,-dx,dy,nx,ny,nz)

        # Loop over all hour angle weights
        for hour_angle in hour_angles:

            # Compute temporal shift in degrees
            shift = hour_angle['angle']

            # Initialise shifted zenith angle map
            zmap_shift = gammalib.GSkyMap('CAR','CEL',shift,0.0,-dx,dy,nx,ny)

            # Merge zenith angle map into shifted map
            zmap_shift += zmap

            # Loop over all pixels of the shifted map and add the hours during
            # which the shifted map occurs to the relevant zenith angle bin
            for i, zenith in enumerate(zmap_shift):
                iz = int(zenith/dz)
                if iz < nz:
                    self._cube[i,iz] += hour_angle['hours']

        # Return
        return

    def _save_results(self, outfile, clobber):
        """
        Save results in VISIBILITY FITS table

        Parameters
        ----------
        outfile : str
            Result FITS file name
        clobber : bool
            Overwrite existing file?
        """
        # Create FITS table columns
        nrows         = len(self._results)
        time          = gammalib.GFitsTableStringCol('Time', nrows, 20)
        mjd           = gammalib.GFitsTableDoubleCol('MJD', nrows)
        sun_ra        = gammalib.GFitsTableDoubleCol('RA_sun', nrows)
        sun_dec       = gammalib.GFitsTableDoubleCol('DEC_sun', nrows)
        moon_ra       = gammalib.GFitsTableDoubleCol('RA_moon', nrows)
        moon_dec      = gammalib.GFitsTableDoubleCol('DEC_moon', nrows)
        sun_ra_start  = gammalib.GFitsTableDoubleCol('RA_sun_start', nrows)
        sun_ra_stop   = gammalib.GFitsTableDoubleCol('RA_sun_stop', nrows)
        moon_ra_start = gammalib.GFitsTableDoubleCol('RA_moon_start', nrows)
        moon_ra_stop  = gammalib.GFitsTableDoubleCol('RA_moon_stop', nrows)
        elongation    = gammalib.GFitsTableDoubleCol('Elongation', nrows)
        fli           = gammalib.GFitsTableDoubleCol('FLI', nrows)
        dark_time     = gammalib.GFitsTableDoubleCol('Darktime', nrows)

        # Set units of table columns
        mjd.unit('days')
        sun_ra.unit('deg')
        sun_dec.unit('deg')
        moon_ra.unit('deg')
        moon_dec.unit('deg')
        sun_ra_start.unit('deg')
        sun_ra_stop.unit('deg')
        moon_ra_start.unit('deg')
        moon_ra_stop.unit('deg')
        elongation.unit('deg')
        dark_time.unit('hours')

        # File FITS table columns
        for i, result in enumerate(self._results):
            time[i]          = result['time'].utc()
            mjd[i]           = result['time'].mjd()
            sun_ra[i]        = result['sun_ra']
            sun_dec[i]       = result['sun_dec']
            moon_ra[i]       = result['moon_ra']
            moon_dec[i]      = result['moon_dec']
            sun_ra_start[i]  = result['sun_ra_start']
            sun_ra_stop[i]   = result['sun_ra_stop']
            moon_ra_start[i] = result['moon_ra_start']
            moon_ra_stop[i]  = result['moon_ra_stop']
            elongation[i]    = result['elongation']
            fli[i]           = result['fli']
            dark_time[i]     = result['dark_time']

        # Initialise FITS Table with extension "VISIBILITY"
        table = gammalib.GFitsBinTable(nrows)
        table.extname('VISIBILITY')

        # Add Header for compatibility with gammalib.GMWLSpectrum
        table.card('INSTRUME', 'CTA', 'Name of Instrument')
        table.card('TELESCOP', 'CTA', 'Name of Telescope')

        # Append filled columns to fits table    
        table.append(time)
        table.append(mjd)
        table.append(sun_ra)
        table.append(sun_dec)
        table.append(moon_ra)
        table.append(moon_dec)
        table.append(sun_ra_start)
        table.append(sun_ra_stop)
        table.append(moon_ra_start)
        table.append(moon_ra_stop)
        table.append(elongation)
        table.append(fli)
        table.append(dark_time)

        # Append table to result FITS file
        fits = gammalib.GFits(outfile)
        fits.append(table)
        fits.save(clobber)

        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Compute visibility cube
        self._visibility_cube()

        # Optionally publish map
        if self['publish'].boolean():
            self.publish()

        # Return
        return

    def save(self):
        """
        Save the visibility cube
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save visibility cube')

        # Get outfile parameter
        outfile = self['outfile'].filename()

        # Log file name
        self._log_value(gammalib.NORMAL, 'Visibility cube file', outfile.url())

        # Save the visibility cube
        self._cube.save(outfile, self['clobber'].boolean())

        # Save the results in VISIBILITY table extension
        self._save_results(outfile, self['clobber'].boolean())

        # Return
        return

    def publish(self, name=''):
        """
        Publish visibility cube

        Parameters
        ----------
        name : str, optional
            Name of visibility cube
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Publish visibility cube')

        # Set default name is user name is empty
        if not name:
            user_name = self._name()
        else:
            user_name = name

        # Log cube name
        self._log_value(gammalib.TERSE, 'Visibility cube name', user_name)

        # Publish cube
        self._cube.publish(user_name)

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csviscube(sys.argv)

    # Execute application
    app.execute()



#!/usr/bin/env python
# ==========================================================================
# Copies IACT data from remote machine
#
# Copyright (C) 2016-2018 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import sys
import json
import shutil
import gammalib
import ctools

# ================ #
# csiactdata class #
# ================ #
class csiactcopy(ctools.cscript):
    """
    Copies IACT data from remote machine
    
    This script copies IACT data from one location to another. It can
    take a list of observation IDs to allow the download specific
    observations. Index files get merged and updated accordingly.
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Set name
        self._datapath      = os.getenv('VHEFITS','')
        self._remote_master = gammalib.GFilename()
        self._remote_base   = ''
        self._prodname      = ''
        self._outpath       = ''
        self._runlist       = gammalib.GFilename()
        self._runs          = []

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile and setup the observation
        """
        # Get Parameters
        self._remote_master = self['remote_master'].filename()
        if not self._remote_master.exists():
            raise RuntimeError('*** ERROR: FITS data not available. No '
                               'master index file found in "'+
                               self._remote_master+'". Make sure remote '
                               'file system is properly mounted.')
        
        # Get parameters
        self._prodname = self['prodname'].string()
        self._outpath  = gammalib.expand_env(self['outpath'].string())
        if self['runlist'].is_valid():
            self._runlist = self['runlist'].filename()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)
        
        # Return
        return

    def _copy(self, source, clobber):     
        """
        Copy file to self._outpath directory
        
        Parameters
        ----------
        source : str
            Path of file to be copied

        Returns
        -------
        filesize : float
            Size of the file that was copied
        """
        # Get file destination
        destination = os.path.join(self._outpath,
                                   os.path.relpath(source, self._remote_base))
            
        # Initialise return value
        filesize = 0.0

        # Flag if file destination is already available
        is_file = os.path.isfile(destination)
        
        # Logging
        self._log_header3(gammalib.VERBOSE, 'Copy file')
        self._log_value(gammalib.VERBOSE, 'Source', source)
        self._log_value(gammalib.VERBOSE, 'Destination', destination)
        self._log_value(gammalib.VERBOSE, 'Already exists', is_file)
        self._log_value(gammalib.VERBOSE, 'Overwrite', clobber)
        
        # Check if file could be skipped because clobber=no
        if is_file and clobber == False:
            self._log_value(gammalib.VERBOSE, 'Copying', 'Skip (clobber=no)')
        
        # ... check if file could be skipped because it is the same file as
        # inferred from the file path and name (the file is in fact not
        # checked) ...
        elif is_file and os.path.samefile(destination, source):
            self._log_value(gammalib.VERBOSE, 'Copying', 'Skip (same file)')
        
        # ... otherwise copy file
        else:
            # Create directory if it does not exist
            dest_dir = os.path.dirname(destination)
            if not os.path.isdir(dest_dir):
                os.makedirs(dest_dir)

            # If destination file exists then delete it. Make sure that file
            # is writable for that
            if os.path.isfile(destination):
                os.chmod(destination, 420)
                os.remove(destination)

            # Copy file
            shutil.copy2(source, destination)
            
            # Get filesize
            filesize = os.stat(destination).st_size
            
            # Logging
            self._log_value(gammalib.VERBOSE, 'Copying', 'Done!')
        
        # Return 
        return filesize

    def _merge(self, localfits, remotefits, hduname, clobber):
        """
        Merge remote and local fits files

        If the local fits file is not present, a new one is created.
        
        Parameters
        ----------
        localfits : str
            Path of local index FITS file         
        remotefits : str
            Path of remote index FITS file         
        hduname : str
            Name of HDU extension to be merged       
        clobber : bool
            Flag if remote content should overwrite local content        
        """    
        # Write input parameters into logger
        self._log_value(gammalib.NORMAL, 'Local file', localfits)
        self._log_value(gammalib.NORMAL, 'Remote file', remotefits)
        self._log_value(gammalib.NORMAL, 'HDU name', hduname)
        if clobber:
            preference = 'Remote'
        else:
            preference = 'Local'
        self._log_value(gammalib.NORMAL, 'Conflict preference', preference)
    
        # Check if local fits file is available
        if os.path.isfile(localfits):    
            local = gammalib.GFits(str(localfits))
            local_hdu = local[hduname]
        else:
            # Otherwise load remote fits file and delete rows
            local = gammalib.GFits(str(remotefits))
            local_hdu = local[hduname]
            local_hdu.remove_rows(0,local_hdu.nrows())
        
        # find local obs_id 
        local_obs = []
        lobs_col  = local_hdu['OBS_ID']
        for obs_id in lobs_col:
            local_obs.append(obs_id)      
             
        # Initialise run list. A kluge has been implemented to circumvent a
        # too long file name: the run list is split into chunks of 50
        # observations maximum. Note also the "//" division operator that
        # makes sure that "i" is an integer on Python 3.
        runlists       = []
        runs_per_chunk = 50
        for i in range(len(self._runs) // runs_per_chunk + 1):
            chunk = self._runs[i * runs_per_chunk : (i + 1) * runs_per_chunk]
            runlists.append(chunk)

        # Loop over runlist chunks
        for i, runlist in enumerate(runlists):
        
            # Create selection expression for opening remote fits file
            selection = ''
            if len(runlist):
                selection += '['+hduname+']['
                for run in runlist:
                    selection += 'OBS_ID=='+str(run)
                    selection += '||'
                selection = selection[:-2]
                selection += ']'
        
            # Create file name
            remotefile = remotefits + selection
            
            # Open remote fits file
            remote = gammalib.GFits(str(remotefile))
            
            # If first iteration, create clone of FITS column for later use
            if i == 0:
                remote_hdu = remote[hduname].clone()
            else:
                
                # get temporary hdu to be added to current HDU
                tmp_hdu = remote[hduname]
                
                # Get old size of remote HDU
                size = remote_hdu.nrows()
                
                # Extend remote HDU
                remote_hdu.append_rows(tmp_hdu.nrows())
                
                # Loop over rows and columns to copy over values exteding remote HDU
                for col in tmp_hdu:
                    for row in range(col.nrows()):
                        remote_hdu[col.name()][size + row] = col[row]

        # Get remote obs_id
        remote_obs = []
        robs_col   = remote_hdu['OBS_ID']
        for obs_id in robs_col:
            remote_obs.append(obs_id)
        
        # Log entries of index files   
        self._log_value(gammalib.NORMAL, 'Remote entries', len(remote_obs))
        self._log_value(gammalib.NORMAL, 'Local entries', len(local_obs))

        # initialise counter for logging
        removed_rows = 0
        
        # If clobber=yes, we use the remote content to overwrite
        # corresponding local content
        if clobber:
            
            # Loop over remote obs_ids
            for remote_obs_id in remote_obs:
                
                # Check if remote obs_id is present locally
                if remote_obs_id in local_obs:
                    
                    # Remove local obs_id entries
                    table_has_obsid = True
                    
                    # Loop over obs ids and overwrite with remote content
                    while table_has_obsid:
                        
                        # First remove old rows
                        for i in range(local_hdu.nrows()):
                            if remote_obs_id == local_hdu['OBS_ID'][i]:
                                local_hdu.remove_rows(i,1)
                                removed_rows += 1
                                break
                        
                        # Replace with remote content
                        table_has_obsid = False
                        for i in range(local_hdu.nrows()):
                            if remote_obs_id == local_hdu['OBS_ID'][i]:
                                table_has_obsid = True
                                break
                            
        # If clobber=false, we keep the local content unchanged and just
        # expand by remote content        
        else:
            # Loop over remote obs_ids
            for local_obs_id in local_obs:
                
                # Check if local obs_id is present remotely
                if local_obs_id in remote_obs:
                    
                    # Remove remote obs_id entries
                    table_has_obsid = True
                    while table_has_obsid:
                        for i in range(remote_hdu.nrows()):
                            if local_obs_id == remote_hdu['OBS_ID'][i]:
                                remote_hdu.remove_rows(i,1)
                                removed_rows += 1
                                break
                        table_has_obsid = False
                        for i in range(remote_hdu.nrows()):
                            if local_obs_id == remote_hdu['OBS_ID'][i]:
                                table_has_obsid = True
                                break
    
        # Store number of local rows
        old_local_rows = local_hdu.nrows()
        
        # Add remote HDUs
        local_hdu.insert_rows(old_local_rows, remote_hdu.nrows())
        
        # Log actions
        if clobber:
            what = 'Removed local rows'
        else:
            what = 'Skipped remote rows'
        self._log_value(gammalib.NORMAL, what, removed_rows)

        # Loop over columns 
        for i in range(local_hdu.ncols()):
            
            # Get local and remote columns
            local_col  = local_hdu[i]
            remote_col = remote_hdu[i]    
            
            # Loop over entries and merge
            for j in range(remote_col.nrows()):
                local_col[j+old_local_rows] = remote_col[j]
        
        # Save local fits file
        local.saveto(str(localfits), True)
               
        # Return
        return

    def _set_runs(self, filename):
        """
        Set the run list

        Parameters
        ----------
        filename : `~gammalib.GFilename`
            Run list file name

        Returns
        -------
        runs : list of int
            Run list
        """
        # Write header into logger
        self._log_header1(gammalib.TERSE, 'Set runlist')
        
        # Initialise runlist
        runs = []
        
        # If runlist file exists then open file and extract the run
        # information into the run list
        if filename.exists():

            # Open the runlist file
            runfile = open(filename.url())
            
            # Read runlist file and append all run ID as integers to the
            # run list. Skip any blank or comment lines
            for line in runfile.readlines():
                if len(line) == 0:
                    continue
                if line[0] == '#':
                    continue
                if len(line.split()) > 0:
                    run = int(line.split()[0])
                    runs.append(run)
                    self._log_value(gammalib.EXPLICIT, 'Run %d' % len(runs), run)
            
            # Close runlist file
            runfile.close()   
            
            # Logging
            self._log_value(gammalib.NORMAL, 'Number of observations',
                            len(runs))
        
        # ... otherwise, if the runlist filename is empty then leave the
        # run list empty. This implies copying all available data.
        elif filename.is_empty():
            self._log_string(gammalib.NORMAL, 'Copy all available data')
        
        # ... otherwise raise an exception
        else:
            msg = '*** ERROR: Runlist file "%s" not available' % filename.url()
            raise RuntimeError(msg)
        
        # Return run list
        return runs


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Make destination directory if not available
        if not os.path.isdir(self._outpath):
            os.makedirs(self._outpath)

        # Set run list
        self._runs = self._set_runs(self._runlist)

        # Check for availability of remote master file
        if not self._remote_master.exists():
            raise RuntimeError('*** ERROR: Remote master file "'+
                               self._remote_master+'" does not exist.')
        
        # Retrieve json data from remote master
        json_data = open(self._remote_master.url()).read()
        data      = json.loads(json_data) 
        if not 'datasets' in data:
            raise RuntimeError('*** ERROR: Key "datasets" not available '
                               'in remote master index file "'+
                               self._remote_master+'".')
        
        # Get array of configurations
        configs = data['datasets']
        
        # Get remote paths
        self._remote_base = self._remote_master.path()
                
        # Initialise flag if prod has been found
        has_prod = False
        
        # Initialise file names to be copied
        files = set()
        
        # Write header into logger
        self._log_header2(gammalib.TERSE, 'Loop over remote configs')
        
        # Loop over configs
        for config in configs: 
            
            # Write header into logger. The str() is needed since
            # "config['name']" is <type 'unicode'>, and under Python 2 a
            # string is expected.
            self._log_header2(gammalib.VERBOSE, str(config['name']))
            
            # Check if prodname was found
            if config['name'] == self._prodname:
                
                # Inidicate that prodname has been found
                has_prod = True           
                
                # Build path of index files
                remote_hdu = str(os.path.join(self._remote_base,
                                              config['hduindx']))
                remote_obs = str(os.path.join(self._remote_base,
                                              config['obsindx']))
                
                # Log information
                self._log_header3(gammalib.NORMAL, 'Remote config "'+
                                  self._prodname+'"')
                self._log_value(gammalib.NORMAL, 'HDU index', remote_hdu)
                self._log_value(gammalib.NORMAL, 'Observation index', remote_obs)
                
                # Open remote HDU index file
                fits = gammalib.GFits(str(remote_hdu))
                table = fits['HDU_INDEX']
                
                # Initialise flag if SIZE column is present
                has_size = table.contains('SIZE')
                
                # Initialise file size
                if has_size:
                    cp_size = 0
                
                # Initialise remote observation IDs
                remote_ids = set()
                for row in range(table.nrows()):
                    remote_ids.add(table['OBS_ID'][row])
                
                # Log runs that are not available remotely
                for run in self._runs:
                    
                    # Check for run not in remote data store
                    if not run in remote_ids:
                        msg = ('Skip observation "%s": ID not available '
                               'remotely' % str(run))
                        self._log_string(msg)
                
                # Loop over remote HDU index file
                for row in range(table.nrows()):
                    
                    # Get observation ID
                    obs_id    = table['OBS_ID'][row]
                    file_dir  = table['FILE_DIR'][row]
                    file_name = table['FILE_NAME'][row]
                    
                    # Skip if filename is empty
                    if file_name == '':
                        continue
                    
                    # Check if we need to consider an input runlist
                    if len(self._runs):
                        
                        # Check if obs id is in runlist
                        if obs_id in self._runs:                                    
                            
                            # Get filename
                            fname = os.path.join(os.path.dirname(remote_hdu),
                                                 file_dir, file_name)
                            
                            # Add file
                            oldlen = len(files)
                            files.add(fname)
                            newlen = len(files)
                            
                            # Add file size
                            if has_size and newlen > oldlen:
                                cp_size += table['SIZE'][row]
                        
                    # Otherwise add every file
                    else:
                        # Get filename
                        fname = os.path.join(os.path.dirname(remote_hdu),
                                             file_dir, file_name)
                        
                        # Add file
                        oldlen = len(files)
                        files.add(fname)
                        newlen = len(files)
                        
                        # Add file size
                        if has_size and newlen > oldlen:
                            cp_size += table['SIZE'][row]
                            
                # Log file information
                self._log_header2(gammalib.NORMAL, 'File information')
                self._log_value(gammalib.NORMAL, 'Number of files', len(files))
                if has_size:
                    size = float(cp_size) * 1.0e-6
                    self._log_value(gammalib.NORMAL, 'Size', '%.2f MB' % size)
                self._log_header3(gammalib.VERBOSE, 'File names')
                for filename in files:
                    self._log_string(gammalib.VERBOSE, str(filename)+'\n')

                # Close HDU index file
                fits.close()
                
            # If prodname is not found just log that we skip the config
            else:
                self._log_header3(gammalib.EXPLICIT, 'Skipping config "'+
                                  str(config['name'])+'"')
        
        # Raise Exception if prodname was not found
        if not has_prod: 
            msg = '*** ERROR: FITS production "'+self._prodname+'" not '
            msg += 'available. Available productions are:\n'
            for config in configs:
                msg += ' - '+config['name']+'\n'
            raise RuntimeError(msg)

        # Write header
        self._log_header1(gammalib.NORMAL, 'Copying files')
        
        # Intialise counter
        k = 0
    
        # Initialise values for logging
        last_fraction      =  0.0
        fraction_increment = 20.0
        
        # Use 10% step increase
        if self._logNormal():
            fraction_increment = 10.0
            
        # Use 5% step increase
        if self._logTerse():
            fraction_increment = 5.0
        
        # Use 2% step increase
        if self._logExplicit():
            fraction_increment = 2.0
        
        # Initialise logging properties
        n_copied   = 0
        total_size = 0.0
        
        # Loop over files and copy
        for filename in files:
            
            # Log progress
            fraction = float(k) / float(len(files)) * 100.0
            while fraction > last_fraction:
                
                # Print status of copying procedure
                self._log_value(gammalib.NORMAL, 'Status', '%d %%' %
                                int(last_fraction))
                last_fraction += fraction_increment
            
            # Copy file
            filesize = self._copy(filename, self._clobber())

            # If the filesize is positive then increment the number of copied
            # files and add the size to the total copied filesize
            if filesize > 0.0:
                total_size += filesize
                n_copied   += 1
            
            # Increment file counter
            k += 1
        
        # Logging
        self._log_value(gammalib.NORMAL, 'Status', 'Finished')

        # Logging about index files
        self._log_header1(gammalib.TERSE, 'Updating index files')

        # Build local hdu index file name
        local_hdu = os.path.join(self._outpath,
                                 os.path.relpath(remote_hdu,
                                                 self._remote_base))
        # Build local obs index file name
        local_obs = os.path.join(self._outpath,
                                 os.path.relpath(remote_obs,
                                                 self._remote_base))
        
        # If we have a runlist then merge index file
        if len(self._runs):
            
            # Logging            
            self._log_header3(gammalib.TERSE, 'HDU index')
                
            # Merge remote index files with local files
            self._merge(local_hdu, remote_hdu, 'HDU_INDEX', self._clobber())
            
            # Logging
            self._log_header3(gammalib.TERSE, 'OBS index')
                
            # Merge remote index files with local files
            self._merge(local_obs, remote_obs, 'OBS_INDEX', self._clobber())
            
        else: 
            # If all files were copied, just copy index files too
            self._copy(remote_hdu, self._clobber())
            self._copy(remote_obs, self._clobber())
        
        # Logging
        self._log_header3(gammalib.TERSE, 'Master index file')
        
        # Adding prodname to local master
        localmaster = os.path.join(self._outpath, 'master.json')
        
        # If no local master is found, copy master over first
        if not os.path.isfile(localmaster):
            self._copy(self._remote_master.url(), self._clobber())
            
        # Load local master
        json_data = open(localmaster).read()
        data      = json.loads(json_data) 
        configs   = data['datasets']
        
        # Initialise flag indicating if we already have prodname in master
        # file
        has_config = False
        
        # Initialise new configs array
        newconfigs = []
        
        # Loop over configs in master index file
        for config in configs:
            
            # Get hdu and obs index files
            hdu = os.path.join(self._outpath, config['hduindx'])
            obs = os.path.join(self._outpath, config['obsindx'])
            
            # Check if index files are available
            if not (gammalib.GFilename(str(hdu)).is_fits() and
                    gammalib.GFilename(str(obs)).is_fits()):
                self._log_value(gammalib.NORMAL,
                                'Removing "'+str(config['name']),
                                'Not available')
            else:
                # Append config if valid
                newconfigs.append(config)
                self._log_value(gammalib.NORMAL,
                                'Keeping "'+str(config['name']),
                                'Available')
            
            # Signals that downloaded config is available
            if config['name'] == self._prodname:
                has_config = True
        
        # Create new entry if config was not available    
        if not has_config:
            newdict            = dict.fromkeys(['name','hduindx','obsindx'])
            newdict['name']    = self._prodname
            newdict['hduindx'] = os.path.relpath(local_hdu, self._outpath)
            newdict['obsindx'] = os.path.relpath(local_obs, self._outpath)
            newconfigs.append(newdict)
            if self._logTerse():
                self._log('Adding "'+str(newdict['name'])+'"')
                self._log('\n')
        
        # Write new json master file. Make sure that we have write permission
        # before writing the file. This is needed as the original master file
        # may be read-only, and the shutil.copy2 function copies over the
        # access permissions.
        os.chmod(localmaster, 420)
        f = open(localmaster, 'w')
        data['datasets'] = newconfigs
        json.dump(data, f, indent=2)
        f.close()

        # Compute total size in MB
        size_MB = float(total_size) * 1.0e-6

        # Log summary
        self._log_header1(gammalib.NORMAL, 'Summary')
        self._log_value(gammalib.NORMAL, 'Data files found', len(files))
        self._log_value(gammalib.NORMAL, 'Data files copied', n_copied)
        self._log_value(gammalib.NORMAL, 'Size', '%.2f MB' % size_MB)
        self._log_value(gammalib.NORMAL, 'Local configs', len(newconfigs))
        self._log_header3(gammalib.TERSE, 'Content of master index')
        for config in newconfigs:
            self._log_string(gammalib.TERSE, str(config['name'])+'\n')

        # Return
        return       

# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csiactcopy(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csmodelinfo script.
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import cscripts
from testing import test


# ================================= #
# Test class for csmodelinfo script #
# ================================= #
class Test(test):
    """
    Test class for csmodelinfo script

    This test class makes unit tests for the csmodelinfo script by using it
    from the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set test data
        self._models_spat = self._datadir + '/models_spatial.xml'
        self._models_spec = self._datadir + '/models_spectral.xml'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions.
        """
        # Set test name
        self.name('csmodelinfo')

        # Append tests
        self.append(self._test_cmd, 'Test csmodelinfo on command line')
        self.append(self._test_python, 'Test csmodelinfo from Python')

        # Return
        return

    # Test csmodelinfo on command line
    def _test_cmd(self):
        """
        Test csmodelinfo on the command line
        """
        # Set script name
        csmodelinfo = self._script('csmodelinfo')

        # Setup csmodelinfo command
        cmd = csmodelinfo+' inmodel="'+self._models_spec+'"'+ \
                          ' outds9file="csmodelinfo_cmd1.reg"'+ \
                          ' logfile="csmodelinfo_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check region file
        self._check_ds9_file('csmodelinfo_cmd1.reg')

        # Setup csmodelinfo command
        cmd = csmodelinfo+' inmodel="model_that_does_not_exist.xml"'+ \
                          ' outds9file="csmodelinfo_cmd2.reg"'+ \
                          ' logfile="csmodelinfo_cmd2.log" debug=yes'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csmodelinfo --help
        self._check_help(csmodelinfo)

        # Return
        return

    # Test csmodelinfo from Python
    def _test_python(self):
        """
        Test csmodelinfo from Python
        """
        # Set-up csmodelinfo
        modelinfo = cscripts.csmodelinfo()
        modelinfo['inmodel']    = self._models_spat
        modelinfo['outds9file'] = 'csmodelinfo_py1.reg'
        modelinfo['logfile']    = 'csmodelinfo_py1.log'
        modelinfo['chatter']    = 2

        # Run csmodelinfo script and save DS9 file
        modelinfo.logFileOpen()   # Make sure we get a log file
        modelinfo.run()
        modelinfo.save()

        # Check region file
        self._check_ds9_file('csmodelinfo_py1.reg')

        # Set-up csmodelinfo
        modelinfo = cscripts.csmodelinfo()
        modelinfo['inmodel']    = self._models_spec
        modelinfo['outds9file'] = 'csmodelinfo_py2.reg'
        modelinfo['logfile']    = 'csmodelinfo_py2.log'
        modelinfo['chatter']    = 3

        # Execute csmodelinfo script
        modelinfo.execute()

        # Check region file
        self._check_ds9_file('csmodelinfo_py2.reg')

        # Return
        return

    # Check region file
    def _check_ds9_file(self, filename):
        """
        Check region file
        """
        # Open file   
        f = open(filename,'r')

        # Expect "fk5" in first line
        line = f.readline().strip('\n')
        self.test_assert(line == 'fk5',
                         'Check for "fk5" in first line')

        # Expect "fk5" in first line
        line = f.readline().strip('\n')
        self.test_assert(line[0:6] == 'point(',
                         'Check for "point(" in second line')

        # Close file
        f.close()
        
        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Display phase information from event list
#
# Copyright (C) 2017 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()


# =========== #
# Plot phases #
# =========== #
def plot_phases(filename, plotfile, nphases=40):
    """
    Plot phases

    Parameters
    ----------
    filename : str
        Name of lightcurve FITS file
    plotfile : str
        Plot file name
    """
    # Read observation container. If an exception occurs then try loading the
    # file as an event list and build an observation container.
    try:
        obs = gammalib.GObservations(filename)
    except:
        obs = gammalib.GObservations()
        obs.append(gammalib.GCTAObservation(filename))

    # Initialise phases
    phases = []

    # Loop over all observations
    for run in obs:

        # Get events
        events = run.events()

        # Loop over all events
        for event in events:

            # Get phase
            phase = event.phase()

            # Collect phase
            phases.append(phase)
            phases.append(phase+1.0)

    # Plot phase
    plt.figure()
    plt.hist(phases,bins=nphases,range=(0.0,2.0),facecolor='r')
    plt.xlabel('Phase')
    plt.ylabel('Events')

    # Show figure
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# =========== #
# Show phases #
# =========== #
def show_phases():
    """
    Show phases
    """
    # Set usage string
    usage = 'show_phases.py [-p plotfile] [file]'

    # Set default options
    options = [{'option': '-n', 'value': '20'},
               {'option': '-p', 'value': ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    nphases  = int(options[0]['value'])*2
    plotfile = options[1]['value']

    # Plot phases
    plot_phases(args[0], plotfile, nphases=nphases)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show phases
    show_phases()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctmodel tool.
#
# Copyright (C) 2014-2020 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools
from testing import test


# =========================== #
# Test class for ctmodel tool #
# =========================== #
class Test(test):
    """
    Test class for ctmodel tool
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set test data
        self._obs_unbinned = self._datadir + '/obs_unbinned.xml'
        self._expcube      = self._datadir + '/crab_expcube.fits'
        self._psfcube      = self._datadir + '/crab_psfcube.fits'
        self._edispcube    = self._datadir + '/crab_edispcube.fits'
        self._bkgcube      = self._datadir + '/crab_bkgcube.fits'
        self._bkgmodel     = self._datadir + '/crab_bkgcube.xml'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctmodel')

        # Append tests
        self.append(self._test_cmd, 'Test ctmodel on command line')
        self.append(self._test_python, 'Test ctmodel from Python')

        # Return
        return

    # Test ctmodel on command line
    def _test_cmd(self):
        """
        Test ctmodel on the command line
        """
        # Set tool name
        ctmodel = self._tool('ctmodel')

        # Setup ctmodel command
        cmd = ctmodel+' incube="NONE" inobs="NONE" expcube="NONE"'+\
                      ' psfcube="NONE" bkgcube="NONE"'+ \
                      ' outcube="ctmodel_cmd1.fits"'+ \
                      ' inmodel="'+self._model+'"'+ \
                      ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                      ' rad=5.0 ra=83.63 dec=22.01 tmin=0.0 tmax=1800'+ \
                      ' emin=1.0 emax=100.0 enumbins=5  ebinalg="LOG"'+ \
                      ' nxpix=40 nypix=40'+ \
                      ' binsz=0.1 coordsys="CEL" proj="CAR"'+ \
                      ' xref=83.63 yref=22.01'+ \
                      ' logfile="ctmodel_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check result file
        self._check_result_file('ctmodel_cmd1.fits')

        # Setup ctmodel command
        cmd = ctmodel+' incube="NONE" inobs="NONE" expcube="NONE"'+\
                      ' psfcube="NONE" bkgcube="NONE"'+ \
                      ' outcube="ctmodel_cmd2.fits"'+ \
                      ' inmodel="model_that_does_not_exist.xml"'+ \
                      ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                      ' rad=5.0 ra=83.63 dec=22.01 tmin=0.0 tmax=1800'+ \
                      ' emin=1.0 emax=100.0 enumbins=5  ebinalg="LOG"'+ \
                      ' nxpix=40 nypix=40'+ \
                      ' binsz=0.1 coordsys="CEL" proj="CAR"'+ \
                      ' xref=83.63 yref=22.01'+ \
                      ' logfile="ctmodel_cmd2.log" debug=yes chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check ctmodel --help
        self._check_help(ctmodel)

        # Return
        return

    # Test ctmodel from Python
    def _test_python(self):
        """
        Test ctmodel from Python
        """
        # Allocate ctmodel
        model = ctools.ctmodel()

        # Check that empty ctmodel tool holds an empty model cube
        self._check_cube(model.cube(), nx=0, ny=0, nebins=0)

        # Check that saving does not nothing
        model['outcube']  = 'ctmodel_py0.fits'
        model['logfile']  = 'ctmodel_py0.log'
        model.logFileOpen()
        model.save()
        self.test_assert(not os.path.isfile('ctmodel_py0.fits'),
             'Check that no model cube has been created')

        # Check that clearing does not lead to an exception or segfault
        model.clear()

        # Now set ctmodel parameters
        model['incube']   = 'NONE'
        model['inmodel']  = self._model
        model['inobs']    = 'NONE'
        model['expcube']  = 'NONE'
        model['psfcube']  = 'NONE'
        model['bkgcube']  = 'NONE'
        model['caldb']    = self._caldb
        model['irf']      = self._irf
        model['rad']      = 5
        model['ra']       = 83.63
        model['dec']      = 22.01
        model['tmin']     = 0
        model['tmax']     = 1800
        model['ebinalg']  = 'LOG'
        model['emin']     = 1.0
        model['emax']     = 100.0
        model['enumbins'] = 5
        model['nxpix']    = 40
        model['nypix']    = 40
        model['binsz']    = 0.1
        model['coordsys'] = 'CEL'
        model['proj']     = 'CAR'
        model['xref']     = 83.63
        model['yref']     = 22.01
        model['outcube']  = 'ctmodel_py1.fits'
        model['logfile']  = 'ctmodel_py1.log'
        model['chatter']  = 2

        # Run ctmodel tool
        model.logFileOpen()   # Make sure we get a log file
        model.run()
        model.save()

        # Check result file
        self._check_result_file('ctmodel_py1.fits')

        # Copy ctmodel tool
        cpy_model = model.copy()

        # Check model cube of ctmodel copy
        self._check_cube(cpy_model.cube())

        # Execute copy of ctmodel tool again, now with a higher chatter
        # level than before
        cpy_model['outcube'] = 'ctmodel_py2.fits'
        cpy_model['logfile'] = 'ctmodel_py2.log'
        cpy_model['publish'] = True
        cpy_model['chatter'] = 3
        cpy_model.logFileOpen()  # Needed to get a new log file
        cpy_model.execute()

        # Check result file
        self._check_result_file('ctmodel_py2.fits')

        # Now clear copy of ctmodel tool
        cpy_model.clear()

        # Check that the cleared copy has also cleared the model cube
        self._check_cube(cpy_model.cube(), nx=0, ny=0, nebins=0)

        # Get mixed observation container
        obs = self._obs_mixed()

        # Setup ctmodel tool from observation container and get model cube
        # definition from counts cube
        model = ctools.ctmodel(obs)
        model.models(gammalib.GModels(self._model))
        model['incube']   = self._cntcube
        model['expcube']  = 'NONE'
        model['psfcube']  = 'NONE'
        model['bkgcube']  = 'NONE'
        model['caldb']    = self._caldb
        model['irf']      = self._irf
        model['outcube']  = 'ctmodel_py3.fits'
        model['logfile']  = 'ctmodel_py3.log'
        model['chatter']  = 4

        # Execute ctmodel tool
        model.logFileOpen()  # Needed to get a new log file
        model.execute()

        # Check result file
        self._check_result_file('ctmodel_py3.fits', nx=20, ny=20, nebins=5)

        # Publish with name
        model.publish('My model')

        # Setup ctmodel from event list FITS file
        model = ctools.ctmodel()
        model['inobs']     = self._events
        model['incube']    = 'NONE'
        model['inmodel']   = self._model
        model['expcube']   = 'NONE'
        model['psfcube']   = 'NONE'
        model['edispcube'] = 'NONE'
        model['bkgcube']   = 'NONE'
        model['caldb']     = self._caldb
        model['irf']       = self._irf
        model['rad']       = 5
        model['ra']        = 83.63
        model['dec']       = 22.01
        model['tmin']      = 0
        model['tmax']      = 1800
        model['ebinalg']   = 'POW'
        model['emin']      = 1.0
        model['emax']      = 100.0
        model['enumbins']  = 5
        model['ebingamma'] = 1.0
        model['nxpix']     = 40
        model['nypix']     = 40
        model['binsz']     = 0.1
        model['coordsys']  = 'CEL'
        model['proj']      = 'CAR'
        model['xref']      = 83.63
        model['yref']      = 22.01
        model['outcube']   = 'ctmodel_py4.fits'
        model['logfile']   = 'ctmodel_py4.log'
        model['chatter']   = 2

        # Run ctmodel tool
        model.logFileOpen()   # Make sure we get a log file
        model.execute()

        # Check result file
        self._check_result_file('ctmodel_py4.fits')

        # Setup ctmodel from observation definition XML file
        model = ctools.ctmodel()
        model['inobs']     = self._obs_unbinned
        model['incube']    = 'NONE'
        model['inmodel']   = self._model
        model['expcube']   = 'NONE'
        model['psfcube']   = 'NONE'
        model['edispcube'] = 'NONE'
        model['bkgcube']   = 'NONE'
        model['caldb']     = self._caldb
        model['irf']       = self._irf
        model['rad']       = 5
        model['ra']        = 83.63
        model['dec']       = 22.01
        model['tmin']      = 0
        model['tmax']      = 1800
        model['ebinalg']   = 'LOG'
        model['emin']      = 1.0
        model['emax']      = 100.0
        model['enumbins']  = 5
        model['nxpix']     = 40
        model['nypix']     = 40
        model['binsz']     = 0.1
        model['coordsys']  = 'CEL'
        model['proj']      = 'CAR'
        model['xref']      = 83.63
        model['yref']      = 22.01
        model['outcube']   = 'ctmodel_py5.fits'
        model['logfile']   = 'ctmodel_py5.log'
        model['chatter']   = 2

        # Run ctmodel tool
        model.logFileOpen()   # Make sure we get a log file
        model.execute()

        # Check result file
        self._check_result_file('ctmodel_py5.fits')

        # Setup ctmodel for stacked IRF
        model = ctools.ctmodel()
        model['inobs']     = 'NONE'
        model['incube']    = 'NONE'
        model['inmodel']   = self._bkgmodel
        model['expcube']   = self._expcube
        model['psfcube']   = self._psfcube
        model['edispcube'] = 'NONE'
        model['bkgcube']   = self._bkgcube
        model['edisp']     = False
        model['rad']       = 5
        model['ra']        = 83.63
        model['dec']       = 22.01
        model['tmin']      = 0
        model['tmax']      = 1800
        model['ebinalg']   = 'LOG'
        model['emin']      = 1.0
        model['emax']      = 100.0
        model['enumbins']  = 5
        model['nxpix']     = 40
        model['nypix']     = 40
        model['binsz']     = 0.1
        model['coordsys']  = 'CEL'
        model['proj']      = 'CAR'
        model['xref']      = 83.63
        model['yref']      = 22.01
        model['outcube']   = 'ctmodel_py6.fits'
        model['logfile']   = 'ctmodel_py6.log'
        model['chatter']   = 2

        # Run ctmodel tool
        model.logFileOpen()   # Make sure we get a log file
        model.execute()

        # Check result file
        self._check_result_file('ctmodel_py6.fits')

        # Setup ctmodel for stacked IRF using energy dispersion
        model = ctools.ctmodel()
        model['inobs']     = 'NONE'
        model['incube']    = 'NONE'
        model['inmodel']   = self._bkgmodel
        model['expcube']   = self._expcube
        model['psfcube']   = self._psfcube
        model['edispcube'] = self._edispcube
        model['bkgcube']   = self._bkgcube
        model['edisp']     = True
        model['rad']       = 5
        model['ra']        = 83.63
        model['dec']       = 22.01
        model['tmin']      = 0
        model['tmax']      = 1800
        model['ebinalg']   = 'LOG'
        model['emin']      = 1.0
        model['emax']      = 100.0
        model['enumbins']  = 5
        model['nxpix']     = 40
        model['nypix']     = 40
        model['binsz']     = 0.1
        model['coordsys']  = 'CEL'
        model['proj']      = 'CAR'
        model['xref']      = 83.63
        model['yref']      = 22.01
        model['outcube']   = 'ctmodel_py7.fits'
        model['logfile']   = 'ctmodel_py7.log'
        model['chatter']   = 2

        # Run ctmodel tool
        model.logFileOpen()   # Make sure we get a log file
        model.execute()

        # Check result file
        self._check_result_file('ctmodel_py7.fits')

        # Return
        return

    # Check result file
    def _check_result_file(self, filename, nx=40, ny=40, nebins=5):
        """
        Check result file

        Parameters
        ----------
        filename : str
            Model cube file name
        nx : int, optional
            Number of pixels in X direction
        ny : int, optional
            Number of pixels in Y direction
        nebins : int, optional
            Number of energy bins
        """
        # Open result file
        cube = gammalib.GCTAEventCube(filename)

        # Check cube
        self._check_cube(cube, nx=nx, ny=ny, nebins=nebins)

        # Return
        return

    # Check model cube
    def _check_cube(self, cube, nx=40, ny=40, nebins=5):
        """
        Check model cube

        Parameters
        ----------
        cube : `~gammalib.GCTAEventCube`
            Model cube
        nx : int, optional
            Number of pixels in X direction
        ny : int, optional
            Number of pixels in Y direction
        nebins : int, optional
            Number of energy bins
        """
        # Check number of bins per axis
        self.test_value(cube.nx(), nx, 'Check number of X pixels')
        self.test_value(cube.ny(), ny, 'Check number of Y pixels')
        self.test_value(cube.ebins(), nebins, 'Check number of energy bins')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Shows source variability evolution created with ctfindvar
#
# Copyright (C) 2018 Simon Bonnefoy
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()


# ========================== #
# Plot variability evolution #
# ========================== #
def plot_variability_evolution(filename, plotfile, srcname):
    """
    Plot variability evolution 

    Parameters
    ----------
    filename : str
        Name of spectrum FITS file
    plotfile : str
        Plot file name
    srcname : str
        Source name
    """
    # Read variability  file    
    fits    = gammalib.GFits(filename)
    table   = fits.table(1)
    c_gti_start  = table['TSTART']
    c_gti_stop   = table['TSTOP']
    c_sig        = table[srcname]

    # Convert the c_sig object to list
    table_c_sig = []
    table_c_gti = []
    for k in c_sig:
        table_c_sig.append(k)
    for j, k in zip(c_gti_start, c_gti_stop):
        table_c_gti.append(0.5*(j+k))

    # Initializing the plots
    axarr = []
    f     = plt.figure(figsize=(12,7))
    ax1   = f.add_subplot(121)
    ax2   = f.add_subplot(122)
    axarr.append(ax1)
    axarr.append(ax2)

    # Create first plot
    axarr[0].plot(table_c_gti, table_c_sig)
    axarr[0].set_title('Significance evolution')
    axarr[0].set_xlabel('MJD')
    axarr[0].set_ylabel('Significance')

    # Create second plot
    axarr[1].hist(table_c_sig, 20, range=[min(table_c_sig), max(table_c_sig)],
                  histtype='stepfilled', facecolor='g', alpha=0.75)
    axarr[1].set_title('Significance distribution')
    axarr[1].set_xlabel('Significance')
    axarr[1].set_ylabel('Counts')
    
    # Optionally save result in file
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ========================== #
# Show variability evolution #
# ========================== #
def show_variability_evolution():
    """
    Show variability 
    """
    # Set usage string
    usage = 'show_variability_evolution.py [-p plotfile] [-s source_name] [file]'

    # Set default options
    options = [{'option': '-p', 'value': ''},
               {'option': '-s', 'value': 'MAXSIGPIXEL'}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    plotfile = options[0]['value']
    srcname  = options[1]['value']

    # Show variability evolution
    plot_variability_evolution(args[0], plotfile, srcname)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show variability evolution 
    show_variability_evolution()
